#include "vale.h"

void finalize();
void buildList();
void getBreakValues();
void matchPointers();
void rewritePointers();


unsigned char fileBuffer[8192];

unsigned int* addresses;
int* pointers;
int addressesCount, pointersCount;

long int ptrTableStart, ptrTableEnd, start, end;
unsigned char* breakValues;
int* breakSequenceLengths;
int breakValuesCount;

int filePtr = -1;

void main(){
  char fileName[30];
  int c;

  while(filePtr == -1){
    printf("Unchanged ROM: ");
    getString(29, fileName);

    filePtr = openFile(fileName, 'r');
    if(filePtr == -1) printf("Unable to open %s.\n", fileName);
    }

  printf("Unchanged ROM text start: ");
  start = hexInput();
  printf("Unchanged ROM text end: ");
  end = hexInput();

  getBreakValues();

  printf("Finding original message-start addresses.\n");
  buildList();
  printf("Found %d message-starts.  Done.\n\n", addressesCount);

  printf("Pointer table start: ");
  ptrTableStart = hexInput();
  printf("Pointer table end: ");
  ptrTableEnd = hexInput();

  printf("Matching pointers to original addresses.\n");
  matchPointers();
  printf("Found %d pointers.  Done.\n\n", pointersCount);

  free(addresses);
  addressesCount = 1;
  close(filePtr);
  filePtr = -1;

  while(filePtr == -1){
    printf("Changed ROM: ");
    getString(29, fileName);

    filePtr = openFile(fileName, 'r');
    if(filePtr == -1) printf("Unable to open %s.\n", fileName);
    else{
      close(filePtr);
      filePtr = openFile(fileName, 'w');
      }
    }

  printf("Changed ROM text start: ");
  start = hexInput();
  printf("Changed ROM text end: ");
  end = hexInput();

  printf("Finding new message-start addresses.\n");
  buildList();
  printf("Found %d message-starts.  Done.\n\n", addressesCount);

  printf("Rewriting pointer table in %s.\n", fileName);
  rewritePointers();
  printf("Done.\n");

  finalize();
  }

void rewritePointers(){
  int c;
  long int currentPos = ptrTableStart;

  lseek(filePtr, ptrTableStart, 0);

  for(c = 0; c < pointersCount; c++){
    if(pointers[c] == -1){ // unmatched pointer
      printf("%lX: can't rewrite pointer (originally unmatched pointer)\n", currentPos);
      currentPos += 2;
      lseek(filePtr, currentPos, 0);
      continue;
      }
    if(pointers[c] >= addressesCount){ // out of bounds
      printf("%lX: can't rewrite pointer (message %d not found in given area of new ROM)\n", currentPos,
          pointers[c] + 1);
      currentPos += 2;
      lseek(filePtr, currentPos, 0);
      continue;
      }
    _write(filePtr, &addresses[pointers[c]], 2);
    currentPos += 2;
    }
  }

void matchPointers(){
  long int currentPos = ptrTableStart;
  unsigned int ptrBuffer;
  int c;

  lseek(filePtr, ptrTableStart, 0);
  pointers = (int*) malloc(sizeof(int));
  pointersCount = 0;

  while(currentPos <= ptrTableEnd){
    pointersCount++;
    pointers = (int*) realloc(pointers, sizeof(int) * pointersCount);

    _read(filePtr, &ptrBuffer, 2);
    //printf("ptrBuffer = %X\n", ptrBuffer);

    for(c = 0; c < addressesCount; c++){
      if(ptrBuffer == addresses[c]) break;
      }
    pointers[pointersCount-1] = c; // pointers holds id# of msg-start add.
    if(pointers[pointersCount-1] == addressesCount){
      printf("%lX: Unmatched pointer warning\n", currentPos);
      pointers[pointersCount-1] = -1;
      }

    currentPos += 2;
    }
  }

void buildList(){
  long int currentPos = start;
  int currentBufferLength;
  int c, d;

  lseek(filePtr, start, 0);
  addresses = (unsigned int*) malloc(sizeof(unsigned int));
  addresses[0] = ((start - 0x10) % 0x4000) + 0x8000;
  addressesCount = 1;
  //printf("addresses[0] = %X\n", addresses[0]);

  while(currentPos <= end){

    if(currentPos + 8192 <= end){ // load next block
      _read(filePtr, fileBuffer, 8192);
      currentBufferLength = 8192;
      }
    else{
      _read(filePtr, fileBuffer, end - currentPos + 1);
      currentBufferLength = end - currentPos + 1;
      }

    //printf("Now checking %lX.\n", currentPos);
    //printf("buffer 0: %X\n", fileBuffer[0]);
    //printf("buffer %X: %X\n", currentBufferLength - 1, fileBuffer[currentBufferLength - 1]);

    c = 0;
    while(c < currentBufferLength){
      for(d = 0; d < breakValuesCount; d++){
        if(fileBuffer[c] == breakValues[d]){
    	  //printf("Break value at %lX. ", currentPos + c);
	  //printf("New message begins at %lX.\n", currentPos + c + breakSequenceLengths[d]);
	  addressesCount++;
	  addresses = (unsigned int*) realloc(addresses, sizeof(unsigned int) * addressesCount);
	  addresses[addressesCount-1] = ((currentPos + c + breakSequenceLengths[d] - 0x10) % 0x4000)
	      + 0x8000;
	  }
        }
      c++;
      }
    currentPos += currentBufferLength;
    //printf("currentPos = %lX.\n", currentPos);
    }
  //for(c = 0; c < addressesCount; c++){
    //printf("Message [%d] start: %X\n", c, addresses[c]);
    //}
  }

void getBreakValues(){
  int c;

  printf("\nNumber of break values: ");
  breakValuesCount = decInput();
  breakValues = (unsigned char*) malloc(sizeof(unsigned char) * breakValuesCount);
  breakSequenceLengths = (int*) malloc(sizeof(int) * breakValuesCount);

  for(c = 0; c < breakValuesCount; c++){
    printf("Break value %d: ", c + 1);
    breakValues[c] = hexInput();
    //printf("breakValue[%d] = %X", c, breakValues[c]);
    printf("Break sequence %d length: ", c + 1);
    breakSequenceLengths[c] = decInput();
    }
  }

void finalize(){
  free(addresses);
  free(pointers);
  free(breakValues);
  free(breakSequenceLengths);
  close(filePtr);
  }